
//WndTimer.cpp

#include "stdafx.h"
#include "WndTimer.h"

#define UWM_TIMER_MSG _T("UWM_TIMER_MSG-{F57C90D5_EC7C_158A_AD22_015F293A5FC0}")

//User Defined Timer Message
const UINT CWndTimer::UWM_TIMER = ::RegisterWindowMessage(UWM_TIMER_MSG);

//Critical Section for protecting the access to the static members
CRITICAL_SECTION CWndTimer::sm_CS;
bool CWndTimer::sm_bIniCS = false;
//Data Map
multimap<HWND, CWndTimer*> CWndTimer::sm_oMultiMap;
//Data Queue
deque<CWndTimer*> CWndTimer::sm_oQueue;
//Thread Handle
HANDLE CWndTimer::sm_hThread = NULL;
//Event Handle
HANDLE CWndTimer::sm_hEvent = NULL;
//Thread Run Flag
bool CWndTimer::sm_bThreadRunning = false;

//CONSTRUCTOR
CWndTimer::CWndTimer() : m_bCreated(false), m_bRunning(false), m_pTimerProc(NULL), 
  m_pArg(NULL), m_uiID(-1), m_iType(-1)
{
  if(false == sm_bIniCS)
  {
    //First Object
    //The Critical section needs Initialization
    ::InitializeCriticalSection(&sm_CS);
    sm_bIniCS = true;
  }
}

//DESTRUCTOR
CWndTimer::~CWndTimer()
{
  if(false == m_bCreated)
    //Nothing to Destroy
    return;
  ::EnterCriticalSection(&sm_CS);
  //=======================================================================
  ASSERT(sm_hThread != NULL);
  if(TYPE_WIN == m_iType)
  {
    multimap<HWND, CWndTimer*>::iterator it=sm_oMultiMap.begin(), itEnd=sm_oMultiMap.end();
    while(it != itEnd)
    {
      if(it->second == this)
      {
        sm_oMultiMap.erase(it);
        break;
      }
      it++;
    }
  }
  else //TYPE_PROC == m_iType
  {
    deque<CWndTimer*>::iterator it1=sm_oQueue.begin(), itEnd1=sm_oQueue.end();
    while(it1 != itEnd1)
    {
      if(*it1 == this)
      {
        sm_oQueue.erase(it1);
        break;
      }
      it1++;
    }
  }
  //=======================================================================
  ::LeaveCriticalSection(&sm_CS);
  if(true==sm_oMultiMap.empty() && true==sm_oQueue.empty())
  {
    //Last Object
    ::DeleteCriticalSection(&sm_CS);
    sm_bIniCS = false;
  }
}

//Creation Functions
int CWndTimer::Create(HWND const& rhWnd, UINT uiID, UINT uiPeriod)
{
  if(true == m_bCreated)
    //Already Created
    return m_uiID;
  //Check the Window
  if((NULL==rhWnd) || (FALSE==::IsWindow(rhWnd)))
    //Error
    return -1;
  ::EnterCriticalSection(&sm_CS);
  //=======================================================================
  m_iType = TYPE_WIN;
  m_uiPeriod = uiPeriod;
  //Check if the ID is already taken
  multimap<HWND, CWndTimer*>::iterator it = sm_oMultiMap.find(rhWnd);
  if(it != sm_oMultiMap.end())
  {
    UINT uiMax = 0;
    bool bFound = false;
    while(it != sm_oMultiMap.upper_bound(rhWnd))
    {
      if(false == bFound)
      {
        if(it->second->m_uiID == uiID)
          bFound = true;
      }
      if(it->second->m_uiID > uiMax)
        uiMax = it->second->m_uiID;
      it++;
    }
    if(true == bFound)
      m_uiID = uiMax+1;
    else
      m_uiID = uiID;
  }
  else
    m_uiID = uiID;
  //Save in Map
  sm_oMultiMap.insert(pair<HWND, CWndTimer*>(rhWnd, this));
  if(NULL == sm_hThread)
  {
    //First object is creating the Thread in Suspended State
    DWORD dwThreadID;
    sm_hThread = ::CreateThread((LPSECURITY_ATTRIBUTES)NULL, (DWORD)0,
      (LPTHREAD_START_ROUTINE)TimerThreadProc, NULL, (DWORD)CREATE_SUSPENDED, (LPDWORD)&dwThreadID);
    ASSERT(sm_hThread != NULL);
    //Create the Event (the system closes the handle automatically when the process terminates)
    sm_hEvent = ::CreateEvent(NULL, FALSE, FALSE, NULL);
    ASSERT(sm_hEvent != NULL);
    //Initialized to Above Normal Priority to ensure higher priority than the User Interface threads
    ::SetThreadPriority(sm_hThread, THREAD_PRIORITY_ABOVE_NORMAL);
    //Start the Thread
    ::ResumeThread(sm_hThread);
  }
  //=======================================================================
  ::LeaveCriticalSection(&sm_CS);
  m_bCreated = true;
  return m_uiID;
}

bool CWndTimer::Create(PTIMERPROC pTimerProc, LPVOID pArg, UINT uiPeriod)
{
  if(true == m_bCreated)
    //Already Created
  return false;
  ::EnterCriticalSection(&sm_CS);
  //=======================================================================
  m_iType = TYPE_PROC;
  m_uiPeriod = uiPeriod;
  m_pTimerProc = pTimerProc;
  m_pArg = pArg;
  //Save in Queue
  sm_oQueue.push_back(this);
  if(NULL == sm_hThread)
  {
    //First object is creating the Thread in Suspended State
    DWORD dwThreadID;
    sm_hThread = ::CreateThread((LPSECURITY_ATTRIBUTES)NULL, (DWORD)0,
      (LPTHREAD_START_ROUTINE)TimerThreadProc, NULL, (DWORD)CREATE_SUSPENDED, (LPDWORD)&dwThreadID);
    ASSERT(sm_hThread != NULL);
    //Create the Event (the system closes the handle automatically when the process terminates)
    sm_hEvent = ::CreateEvent(NULL, FALSE, FALSE, NULL);
    ASSERT(sm_hEvent != NULL);
    //Initialized to Above Normal Priority to ensure higher priority than the User Interface threads
    ::SetThreadPriority(sm_hThread, THREAD_PRIORITY_ABOVE_NORMAL);
    //Start the Thread
    ::ResumeThread(sm_hThread);
  }
  //=======================================================================
  ::LeaveCriticalSection(&sm_CS);
  m_bCreated = true;
  return true;
}

//Starting
bool CWndTimer::Start()
{
  if(false == m_bCreated)
    //Not Created
    return false;
  ::EnterCriticalSection(&sm_CS);
  //=======================================================================
  if(false == m_bRunning)
  {
    //Next Scheduled Tick
    m_bRunning = true;
    m_uiTickNext = ::GetTickCount() + m_uiPeriod;
    //First Tick
    if(TYPE_WIN == m_iType)
    {
      multimap<HWND, CWndTimer*>::iterator it=sm_oMultiMap.begin(), itEnd=sm_oMultiMap.end();
      while(it != itEnd)
      {
        if(it->second == this)
        {
          ::PostMessage(it->first, UWM_TIMER, m_uiID, ::GetTickCount());
          break;
        }
        it++;
      }
    }
    else //TYPE_PROC == m_iType
      m_pTimerProc(m_pArg);
    if(false == sm_bThreadRunning)
    {
      //Resume the thread
      sm_bThreadRunning = true;
      ::LeaveCriticalSection(&sm_CS);
      ::SetEvent(sm_hEvent);
      return true;
    }
  }
  //=======================================================================
  ::LeaveCriticalSection(&sm_CS);
  return true;
}

//Stopping
bool CWndTimer::Stop()
{
  if(false == m_bCreated)
    //Not Created
    return false;
  ::EnterCriticalSection(&sm_CS);
  //=======================================================================
  m_bRunning = false;
  //=======================================================================
  ::LeaveCriticalSection(&sm_CS);
  return true;
}

//Period Setter
void CWndTimer::SetPeriod(UINT uiPeriod)
{
  m_uiPeriod = uiPeriod;
  ::EnterCriticalSection(&sm_CS);
  //=======================================================================
  UINT uiTickNext = ::GetTickCount() + m_uiPeriod;
  if(uiTickNext < m_uiTickNext)
    m_uiTickNext = uiTickNext;
  //=======================================================================
  ::LeaveCriticalSection(&sm_CS);
}

//Delete the Timers of the given Window
bool CWndTimer::Delete(HWND const& rhWnd)
{
  //Check if the Window is valid
  if((rhWnd!=NULL) && ::IsWindow(rhWnd))
  {
    multimap<HWND, CWndTimer*>::iterator it, itEnd;
    ::EnterCriticalSection(&sm_CS);
    //=======================================================================
    it=CWndTimer::sm_oMultiMap.find(rhWnd);
    if(it!=sm_oMultiMap.end())
    {
      itEnd = CWndTimer::sm_oMultiMap.upper_bound(it->first);
      while(it != itEnd)
        it = CWndTimer::sm_oMultiMap.erase(it);
      //=======================================================================
      ::LeaveCriticalSection(&sm_CS);
      //OK, deleted
      return true;
    }
    //=======================================================================
    ::LeaveCriticalSection(&sm_CS);
    //Cannot Find Window
    return false;
  }
  //Invalid Window
  return false;
}

//Thread Function
//The Timer's resolution is imposed to about 10 ms. The Timer's Period should be > 20ms, but for
//good results is recommended >= 100 ms.

#define MAX_TICK  200
#define MIN_TICK  20
#define MIN_TICK2 10

UINT TimerThreadProc(LPVOID pParam)
{
  DWORD dwTick;
  bool bSuspend;
  multimap<HWND, CWndTimer*>::iterator it, itEnd, _itEnd;
  deque<CWndTimer*>::iterator it1, itEnd1;
  while(TRUE)
  {
    ::EnterCriticalSection(&CWndTimer::sm_CS);
    //=======================================================================
    it = CWndTimer::sm_oMultiMap.begin();
    itEnd = CWndTimer::sm_oMultiMap.end();
    while(it != itEnd)
    {
      if(true == it->second->m_bRunning)
      {
        dwTick = ::GetTickCount();
        if(dwTick + MIN_TICK2 >= it->second->m_uiTickNext)
        {
          //Check if the Window is still valid
          if(::IsWindow(it->first))
          {
            //Reschedule Tick
            it->second->m_uiTickNext += it->second->m_uiPeriod;
            //Current Tick
            ::PostMessage(it->first, CWndTimer::UWM_TIMER, it->second->m_uiID, dwTick);
          }
          else
          {
            //Destroy everything related to that Window from the Map
            _itEnd = CWndTimer::sm_oMultiMap.upper_bound(it->first);
            while(it != _itEnd)
              it = CWndTimer::sm_oMultiMap.erase(it);
            continue;
          }
        }
      }
      it++;
    }
    //
    it1 = CWndTimer::sm_oQueue.begin();
    itEnd1 = CWndTimer::sm_oQueue.end();
    while(it1 != itEnd1)
    {
      if(true == (*it1)->m_bRunning)
      {
        dwTick = ::GetTickCount();
        if(dwTick + MIN_TICK2 >= (*it1)->m_uiTickNext)
        {
          //Reschedule Tick
          (*it1)->m_uiTickNext += (*it1)->m_uiPeriod;
          //Current Tick - call the Timer Procedure
          (*it1)->m_pTimerProc((*it1)->m_pArg);
        }
      }
      it1++;
    }
    //
    bSuspend = true;
    //Schedule the Next Global Tick
    dwTick = UINT_MAX; //4294967295 or 0xffffffff
    if(CWndTimer::sm_oMultiMap.size() > 0)
    {	
      for(it=CWndTimer::sm_oMultiMap.begin(); it!=CWndTimer::sm_oMultiMap.end(); it++)
      {
        if(true == it->second->m_bRunning)
        {
          bSuspend = false;
          if(it->second->m_uiTickNext < dwTick)
            dwTick = it->second->m_uiTickNext;
        }
      }
    }
    //
    if(CWndTimer::sm_oQueue.size() > 0)
    {	
      for(it1=CWndTimer::sm_oQueue.begin(); it1!=CWndTimer::sm_oQueue.end(); it1++)
      {
        if(true == (*it1)->m_bRunning)
        {
          bSuspend = false;
          if((*it1)->m_uiTickNext < dwTick)
            dwTick = (*it1)->m_uiTickNext;
        }
      }
    }
    if(false == bSuspend)
    {
      if(dwTick > ::GetTickCount())
      {
        dwTick -= ::GetTickCount();
        if(dwTick < MIN_TICK)
          dwTick = MIN_TICK;
      }
      else
        dwTick = MIN_TICK;
    }
    else //true == bSuspend
    {
      //Suspend the Thread
      CWndTimer::sm_bThreadRunning = false;
      ::LeaveCriticalSection(&CWndTimer::sm_CS);
      ::WaitForSingleObject(CWndTimer::sm_hEvent, INFINITE);
      continue;
    }
    //=======================================================================
    ::LeaveCriticalSection(&CWndTimer::sm_CS);
    //Check at most every MAX_TICK ms because some Timer Periods could be changed
    if(dwTick > MAX_TICK)
      dwTick = MAX_TICK;
    //Sleep until the Next Event
    ::Sleep(dwTick);
  }
  return 0;
}

